from smriprep.workflows.outputs import _bids_relative
from argparse import ArgumentParser
from copy import deepcopy

from bids import BIDSLayout
from nipype import Workflow
import os

import sys
import os
from collections import OrderedDict
from copy import deepcopy
from niworkflows.utils.misc import fix_multi_T1w_source_name

from nipype import __version__ as nipype_ver
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu

from niworkflows.engine.workflows import LiterateWorkflow as Workflow
from niworkflows.interfaces.bids import (
    BIDSInfo
)
from ..utils.bids import collect_data, BIDSPlusDataGrabber

from ..workflows.dfa import init_dfa_workflow

from ..interfaces import SubjectSummary, AboutSummary, DerivativesDataSink
from ..__about__ import __version__

DEFAULT_MEMORY_MIN_GB=0.01

def init_datasink_wf(bids_root: str, output_dir: str, name='datasink_wf'):
    workflow = Workflow(name=name)

    inputnode = pe.Node(niu.IdentityInterface(fields=[
        'dfa_h', 'dfa_ci', 'dfa_rsquared', 'source_file']),
        name='inputnode')

    raw_sources = pe.Node(niu.Function(function=_bids_relative), name='raw_sources')
    raw_sources.inputs.bids_root = bids_root

    ds_dfa_h = pe.Node(DerivativesDataSink(
        base_directory=output_dir, desc='dfa', suffix='hurst'),
        name="ds_dfa_h", run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB)
    workflow.connect([
        (inputnode, raw_sources, [('source_file', 'in_files')]),
        (inputnode, ds_dfa_h, [('source_file', 'source_file'),
                                   ('dfa_h', 'in_file'),
                                   # ('dfa_metadata', 'meta_dict')]),
        # (raw_sources, ds_dfa_h, [('out', 'RawSources')]),
                               ])
    ])

    ds_dfa_ci = pe.Node(DerivativesDataSink(
        base_directory=output_dir, desc='dfa', suffix='ci'),
        name="ds_dfa_ci", run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB)
    workflow.connect([
        (inputnode, ds_dfa_ci, [('source_file', 'source_file'),
                               ('dfa_ci', 'in_file'),
                               ]),
        # (raw_sources, ds_dfa_ci, [('out', 'RawSources')]),
    ])

    ds_dfa_r2 = pe.Node(DerivativesDataSink(
        base_directory=output_dir, desc='dfa', suffix='rsquared'),
        name="ds_dfa_rsquared", run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB)
    workflow.connect([
        (inputnode, ds_dfa_r2, [('source_file', 'source_file'),
                                ('dfa_rsquared', 'in_file'),
                                ]),
        # (raw_sources, ds_dfa_r2, [('out', 'RawSources')]),
    ])

    return workflow



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

# Load modules for compatibility
from niworkflows.interfaces import (
    bids, utils)

from .reports import SubjectSummary, AboutSummary
from .dfa import DFA


class DerivativesDataSink(bids.DerivativesDataSink):
    out_path_base = 'neuroHurst'


__all__ = [
    'bids',
    'utils',
    'SubjectSummary',
    'AboutSummary',
    'DerivativesDataSink',
    'DFA'
]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Class and utilities for testing the workflows module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import unittest
import logging
from networkx.exception import NetworkXUnfeasible

from nipype.pipeline import engine as pe
from nipype.interfaces.base import isdefined
from nipype.interfaces import utility as niu

logging.disable(logging.INFO)  # <- do we really want to do this?


class TestWorkflow(unittest.TestCase):
    ''' Subclass for test within the workflow module.
    invoke tests with ``python -m unittest discover test'''

    def assertIsAlmostExpectedWorkflow(self, expected_name, expected_interfaces,
                                       expected_inputs, expected_outputs,
                                       actual):
        ''' somewhat hacky way to confirm workflows are as expected, but with low confidence '''
        self.assertIsInstance(actual, pe.Workflow)
        self.assertEqual(expected_name, actual.name)

        # assert it has the same nodes
        actual_nodes = [actual.get_node(name)
                        for name in actual.list_node_names()]
        actual_interfaces = [node.interface.__class__.__name__
                             for node in actual_nodes]

        # assert lists equal
        self.assertIsSubsetOfList(expected_interfaces, actual_interfaces)
        self.assertIsSubsetOfList(actual_interfaces, expected_interfaces)

        # assert expected inputs, outputs exist
        actual_inputs, actual_outputs = self.get_inputs_outputs(actual_nodes)

        self.assertIsSubsetOfList(expected_outputs, actual_outputs)
        self.assertIsSubsetOfList(expected_inputs, actual_inputs)

    def assertIsSubsetOfList(self, expecteds, actuals):
        for expected in expecteds:
            self.assertIn(expected, actuals)

    def get_inputs_outputs(self, nodes):
        def get_io_names(pre, ios):
            return [pre + str(io[0]) for io in ios]

        actual_inputs = []
        actual_outputs = []
        node_tuples = [(node.name, node.inputs.items(), node.outputs.items())
                       for node in nodes]
        for name, inputs, outputs in node_tuples:
            pre = str(name) + "."
            actual_inputs += get_io_names(pre, inputs)

            pre = pre if pre[0:-1] != 'inputnode' else ""
            actual_outputs += get_io_names(pre, outputs)

        return actual_inputs, actual_outputs

    def assert_circular(self, workflow, circular_connections):
        ''' check key paths in workflow by specifying some connections that should induce
        circular paths, which trips a NetworkX error.
        circular_connections is a list of tuples:
            [('from_node_name', 'to_node_name', ('from_node.output_field','to_node.input_field'))]
        '''

        for from_node, to_node, fields in circular_connections:
            from_node = workflow.get_node(from_node)
            to_node = workflow.get_node(to_node)
            workflow.connect([(from_node, to_node, fields)])

            self.assertRaises(NetworkXUnfeasible, workflow.write_graph)

            workflow.disconnect([(from_node, to_node, fields)])

    def assert_inputs_set(self, workflow, additional_inputs=None):
        """Check that all mandatory inputs of nodes in the workflow (at the first level) are
        already set. Additionally, check that inputs in additional_inputs are set. An input is
        "set" if it is
            a) defined explicitly (e.g., in the Interface declaration)
            OR
            b) connected to another node's output (e.g., using the workflow.connect method)
        additional_inputs is a dict:
            {'node_name': ['mandatory', 'input', 'fields']}
        """

        additional_inputs = additional_inputs or {}
        dummy_node = pe.Node(niu.IdentityInterface(fields=['dummy']), name='DummyNode')
        node_names = [name for name in workflow.list_node_names() if name.count('.') == 0]
        for node_name in set(node_names + list(additional_inputs.keys())):
            node = workflow.get_node(node_name)
            mandatory_inputs = list(node.inputs.traits(mandatory=True).keys())
            other_inputs = additional_inputs[node_name] if node_name in additional_inputs else []
            for field in set(mandatory_inputs + other_inputs):
                if isdefined(getattr(node.inputs, field)):
                    pass
                else:  # not explicitly defined
                    # maybe it is connected to an output
                    with self.assertRaises(Exception):
                        # throws an error if the input is already connected
                        workflow.connect([(dummy_node, node, [('dummy', field)])])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Base module variables
"""
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__packagename__ = 'neuroHurst'
__copyright__ = 'Copyright 2019, University of Chicago Environmental Neuroscience Laboratory'
__credits__ = ('Contributors: please check the ``.zenodo.json`` file at the top-level folder'
               'of the repository')
__url__ = 'https://github.com/enlberman/hurst_pipeline'

DOWNLOAD_URL = (
    'https://github.com/enlberman/{name}/archive/{ver}.tar.gz'.format(
        name=__packagename__, ver=__version__))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This pipeline is developed by the Environmental Neuroscience Lab at the University of Chicago
(https://enl.uchicago.edu/) for open-source software distribution.
"""

from .__about__ import (  # noqa
    __version__,
    __copyright__,
    __credits__,
    __packagename__,
)

import warnings

# cmp is not used by neuroHurst, so ignore nipype-generated warnings
warnings.filterwarnings('ignore', r'cmp not installed')
warnings.filterwarnings('ignore', r'This has not been fully tested. Please report any failures.')
warnings.filterwarnings('ignore', r"can't resolve package from __spec__ or __package__")
warnings.simplefilter('ignore', DeprecationWarning)
warnings.simplefilter('ignore', ResourceWarning)


# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
fractal scaling calculation workflow
=====
"""

import logging
import sys


logging.addLevelName(25, 'IMPORTANT')  # Add a new level between INFO and WARNING
logging.addLevelName(15, 'VERBOSE')  # Add a new level between INFO and DEBUG
logger = logging.getLogger('cli')


def main():
    """Entry point"""
    from .run_utils import get_workflow
    import sentry_sdk
    from ..utils.bids import write_derivative_description

    errno = 1  # Default is error exit unless otherwise set
    workflow, plugin_settings, opts, output_dir, work_dir, bids_dir, subject_list, run_uuid = get_workflow(logger)

    try:
        workflow.run(**plugin_settings)
    except Exception as e:
        if not opts.notrack:
            from ..utils.sentry import process_crashfile
            crashfolders = [output_dir / 'neuroHurst' / 'sub-{}'.format(s) / 'log' / run_uuid
                            for s in subject_list]
            for crashfolder in crashfolders:
                for crashfile in crashfolder.glob('crash*.*'):
                    process_crashfile(crashfile)

            if "Workflow did not execute cleanly" not in str(e):
                sentry_sdk.capture_exception(e)
        logger.critical('neuroHurst failed: %s', e)
        raise
    else:
        errno = 0
        logger.log(25, 'neuroHurst finished without errors')
        if not opts.notrack:
            sentry_sdk.capture_message('neuroHurst finished without errors',
                                       level='info')
    finally:
        from niworkflows.reports import generate_reports
        from subprocess import check_call, CalledProcessError, TimeoutExpired
        from pkg_resources import resource_filename as pkgrf
        from shutil import copyfile

        citation_files = {
            ext: output_dir / 'neuroHurst' / 'logs' / ('CITATION.%s' % ext)
            for ext in ('bib', 'tex', 'md', 'html')
        }

        if citation_files['md'].exists():
            # Generate HTML file resolving citations
            cmd = ['pandoc', '-s', '--bibliography',
                   pkgrf('neuroHurst', 'data/boilerplate.bib'),
                   '--filter', 'pandoc-citeproc',
                   '--metadata', 'pagetitle="neuroHurst citation boilerplate"',
                   str(citation_files['md']),
                   '-o', str(citation_files['html'])]

            logger.info('Generating an HTML version of the citation boilerplate...')
            try:
                check_call(cmd, timeout=10)
            except (FileNotFoundError, CalledProcessError, TimeoutExpired):
                logger.warning('Could not generate CITATION.html file:\n%s',
                               ' '.join(cmd))

            # Generate LaTex file resolving citations
            cmd = ['pandoc', '-s', '--bibliography',
                   pkgrf('neuroHurst', 'data/boilerplate.bib'),
                   '--natbib', str(citation_files['md']),
                   '-o', str(citation_files['tex'])]
            logger.info('Generating a LaTeX version of the citation boilerplate...')
            try:
                check_call(cmd, timeout=10)
            except (FileNotFoundError, CalledProcessError, TimeoutExpired):
                logger.warning('Could not generate CITATION.tex file:\n%s',
                               ' '.join(cmd))
            else:
                copyfile(pkgrf('neuroHurst', 'data/boilerplate.bib'),
                         citation_files['bib'])
        else:
            logger.warning('neuroHurst could not find the markdown version of '
                           'the citation boilerplate (%s). HTML and LaTeX versions'
                           ' of it will not be available', citation_files['md'])

        # Generate reports phase
        failed_reports = generate_reports(
            subject_list, output_dir, work_dir, run_uuid, packagename='neuroHurst')
        write_derivative_description(bids_dir, output_dir / 'neuroHurst')

        if failed_reports and not opts.notrack:
            sentry_sdk.capture_message(
                'Report generation failed for %d subjects' % failed_reports,
                level='error')
        sys.exit(int((errno + failed_reports) > 0))


if __name__ == '__main__':
    raise RuntimeError("neuroHurst/cli/run.py should not be run directly;\n"
                       "Please `pip install` neuroHurst and use the `neuroHurst` command")



from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec, SimpleInterface,
    File, InputMultiPath, OutputMultiPath)
from nipype import logging
from python_fractal_scaling.dfa import dfa
import pandas
import numpy
import nibabel
import nilearn
import os
from nipype.utils.filemanip import fname_presuffix

LOGGER = logging.getLogger('nipype.interface')


class DFAInputSpec(TraitedSpec):
    bold = traits.File(mandatory=True, desc='input bold')
    brainmask = traits.String( mandatory=False, desc='brainmask in bold space')
    csv = traits.String(mandatory=False, desc='input csv')
    max_frequency = traits.Float(.1, usedefault=True,
                             desc='Maximum frequency')
    min_frequency = traits.Float(.01, usedefault=True,
                             desc='Minimum frequency')
    # output_format = traits.String('', usedefault=True,
    #                              desc='output format: default is the same as input format')

    drop_vols = traits.Int(5, usedefault=True,
                                  desc='how many volumes to drop')


class DFAOutputSpec(TraitedSpec):
    out_report = File(exists=True, desc='conformation report')
    hurst = File(exists=True, desc='hurst file')
    confidence_intervals = File(exists=True, desc='confidence interval file')
    rsquared = File(exists=True, desc='r squared file')


def _bold_native_masked_derivative(bold_img, mask_img, derivative_data, out_file):
    from nilearn.image import index_img
    bold_template = index_img(bold_img, 0)
    template_data = bold_template.get_data()
    mask = mask_img.get_data() == 1
    template_data[~mask] = 0
    template_data[mask] = derivative_data
    bold_template.__class__(template_data, bold_template.affine, bold_template.header).to_filename(out_file)


class DFA(SimpleInterface):
    """
    Finds template target dimensions for a series of T1w images, filtering low-resolution images,
    if necessary.

    Along each axis, the minimum voxel size (zoom) and the maximum number of voxels (shape) are
    found across images.

    The ``max_scale`` parameter sets a bound on the degree of up-sampling performed.
    By default, an image with a voxel size greater than 3x the smallest voxel size
    (calculated separately for each dimension) will be discarded.

    To select images that require no scaling (i.e. all have smallest voxel sizes),
    set ``max_scale=1``.
    """
    input_spec = DFAInputSpec
    output_spec = DFAOutputSpec

    def _run_interface(self, runtime):
        img = nibabel.load(str(self.inputs.bold))
        tr = img.header.get('pixdim')[4]

        use_bold = self.inputs.brainmask is not None

        if use_bold:
            mask_img = nibabel.load(str(self.inputs.brainmask))
            data = nilearn.masking.apply_mask(imgs=img,mask_img=mask_img).T
        else:
            data = pandas.read_csv(str(self.inputs.csv))

        data = data[:,self.inputs.drop_vols:].T

        mn = int(numpy.ceil(1 / (tr * self.inputs.max_frequency)))
        mx = int(numpy.floor(1 / (tr * self.inputs.min_frequency)))

        h, hci, rs = dfa(data, max_window_size=mx, min_widow_size=mn)
        hci = numpy.vstack(hci)

        if use_bold:
            out_hurst = fname_presuffix(self.inputs.bold, suffix='_hurst', newpath=os.getcwd())
            out_cis = fname_presuffix(self.inputs.bold, suffix='_hurst_ci', newpath=os.getcwd())
            out_r2s = fname_presuffix(self.inputs.bold, suffix='_hurst_r2', newpath=os.getcwd())

            _bold_native_masked_derivative(bold_img=img, mask_img=mask_img, derivative_data=h, out_file=out_hurst)
            _bold_native_masked_derivative(bold_img=img, mask_img=mask_img, derivative_data=hci[:,0], out_file=out_cis)
            _bold_native_masked_derivative(bold_img=img, mask_img=mask_img, derivative_data=rs, out_file=out_r2s)
        else:
            out_hurst = fname_presuffix(self.inputs.bold, suffix='_hurst.csv', use_ext=False)
            out_cis = fname_presuffix(self.inputs.bold, suffix='_hurst_ci.csv', use_ext=False)
            out_r2s = fname_presuffix(self.inputs.bold, suffix='_hurst_r2.csv', use_ext=False)

            numpy.savetxt(out_hurst, h, delimiter=',')
            numpy.savetxt(out_cis, hci, delimiter=',')
            numpy.savetxt(out_r2s, rs, delimiter=',')

        self._results['hurst'] = out_hurst
        self._results['confidence_intervals'] = out_cis
        self._results['rsquared'] = out_r2s

        return runtime



import os
from pathlib import Path
import logging
import sys
import gc
import warnings
from argparse import ArgumentParser
from argparse import ArgumentDefaultsHelpFormatter


def _warn_redirect(message, category, filename, lineno, logger, file=None, line=None):
    logger.warning('Captured warning (%s): %s', category, message)


def check_deps(workflow):
    from nipype.utils.filemanip import which
    return sorted(
        (node.interface.__class__.__name__, node.interface._cmd)
        for node in workflow._get_all_nodes()
        if (hasattr(node.interface, '_cmd') and
            which(node.interface._cmd.split()[0]) is None))


def get_parser():
    """Build parser object"""
    from smriprep.cli.utils import ParseTemplates, output_space as _output_space
    # from templateflow.api import templates
    from packaging.version import Version
    from ..__about__ import __version__
    # from ..config import NONSTANDARD_REFERENCES
    from .version import check_latest, is_flagged

    verstr = 'neuroHurst v{}'.format(__version__)
    currentv = Version(__version__)
    is_release = not any((currentv.is_devrelease, currentv.is_prerelease, currentv.is_postrelease))

    parser = ArgumentParser(description='neuroHurst: fractal scaling calculations for neuroimaging',
                            formatter_class=ArgumentDefaultsHelpFormatter)

    # Arguments as specified by BIDS-Apps
    # required, positional arguments
    # IMPORTANT: they must go directly with the parser object
    parser.add_argument('bids_dir', action='store', type=Path,
                        help='the root folder of a BIDS valid dataset (sub-XXXXX folders should '
                             'be found at the top level in this folder).')
    parser.add_argument('output_dir', action='store', type=Path,
                        help='the output path for the outcomes of preprocessing and visual '
                             'reports')
    parser.add_argument('analysis_level', choices=['participant'],
                        help='processing stage to be run, only "participant" in the case of '
                             'neuroHurst (see BIDS-Apps specification).')
    parser.add_argument('source_format', choices=['bold', 'csv'],
                        help='Which file format is the time series data coming from?')
    # optional arguments
    parser.add_argument('--version', action='version', version=verstr)

    parser.add_argument('--algorithm', action='store', choices=['dfa'],
                        help='which fractal scaling algorithm to run')

    parser.add_argument('--minimum_frequency', action='store', type=float, default=0.01, required=False,
                        help='minimim frequency')

    parser.add_argument('--maximum_frequency', action='store', type=float, default=0.1, required=False,
                        help='maximum frequency')

    parser.add_argument('--skip_vols', action='store', type=int, default=5, required=False,
                        help='number of volumes to drop')

    g_bids = parser.add_argument_group('Options for filtering BIDS queries')
    g_bids.add_argument('--skip_bids_validation', '--skip-bids-validation', action='store_true',
                        default=False,
                        help='assume the input dataset is BIDS compliant and skip the validation')
    g_bids.add_argument('--participant_label', '--participant-label', action='store', nargs='+',
                        help='a space delimited list of participant identifiers or a single '
                             'identifier (the sub- prefix can be removed)')
    g_bids.add_argument('-t', '--task-id', action='store',
                        help='select a specific task to be processed')

    g_perfm = parser.add_argument_group('Options to handle performance')
    g_perfm.add_argument('--nthreads', '--n_cpus', '-n-cpus', action='store', type=int,
                         help='maximum number of threads across all processes')
    g_perfm.add_argument('--omp-nthreads', action='store', type=int, default=0,
                         help='maximum number of threads per-process')
    g_perfm.add_argument('--mem_mb', '--mem-mb', action='store', default=0, type=int,
                         help='upper bound memory limit for FMRIPREP processes')
    g_perfm.add_argument('--low-mem', action='store_true',
                         help='attempt to reduce memory usage (will increase disk usage '
                              'in working directory)')
    g_perfm.add_argument('--use-plugin', action='store', default=None,
                         help='nipype plugin configuration file')
    g_perfm.add_argument('--boilerplate', action='store_true',
                         help='generate boilerplate only')
    g_perfm.add_argument("-v", "--verbose", dest="verbose_count", action="count", default=0,
                         help="increases log verbosity for each occurence, debug level is -vvv")
    g_conf = parser.add_argument_group('Workflow configuration')
    # g_conf.add_argument(
    #     '--ignore', required=False, action='store', nargs="+", default=[],
    #     choices=['fieldmaps', 'slicetiming', 'sbref'],
    #     help='ignore selected aspects of the input dataset to disable corresponding '
    #          'parts of the workflow (a space delimited list)')
    g_conf.add_argument(
        '--output-format', required=False, action='store', type=str, nargs='+',
        choices=['nifti', 'csv', 'mat', 'hdf5', 'pkl', 'npy'],
        help='choose an output format. defaults to whatever the input file format is.'
    )
    g_conf.add_argument(
        '--dummy-scans', required=False, action='store', default=None, type=int,
        help='Number of timepoints to drop.')

    g_other = parser.add_argument_group('Other options')
    g_other.add_argument('-w', '--work-dir', action='store', type=Path, default=Path('work'),
                         help='path where intermediate results should be stored')
    g_other.add_argument(
        '--resource-monitor', action='store_true', default=False,
        help='enable Nipype\'s resource monitoring to keep track of memory and CPU usage')
    g_other.add_argument(
        '--reports-only', action='store_true', default=False,
        help='only generate reports, don\'t run workflows. This will only rerun report '
             'aggregation, not reportlet generation for specific nodes.')
    g_other.add_argument(
        '--run-uuid', action='store', default=None,
        help='Specify UUID of previous run, to include error logs in report. '
             'No effect without --reports-only.')
    g_other.add_argument('--write-graph', action='store_true', default=True,
                         help='Write workflow graph.')
    g_other.add_argument('--stop-on-first-crash', action='store_true', default=False,
                         help='Force stopping on first crash, even if a work directory'
                              ' was specified.')
    g_other.add_argument('--notrack', action='store_true', default=True,
                         help='Opt-out of sending tracking information of this run to '
                              'the neuroHurst developers. This information helps to '
                              'improve neuroHurst and provides an indicator of real '
                              'world usage crucial for obtaining funding.')

    latest = check_latest()
    if latest is not None and currentv < latest:
        print("""\
You are using neuroHurst-%s, and a newer version of neuroHurst is available: %s.
Please check out our documentation about how and when to upgrade""" % (
            __version__, latest), file=sys.stderr)

    _blist = is_flagged()
    if _blist[0]:
        _reason = _blist[1] or 'unknown'
        print("""\
WARNING: Version %s of neuroHurst (current) has been FLAGGED
(reason: %s).
That means some severe flaw was found in it and we strongly
discourage its usage.""" % (__version__, _reason), file=sys.stderr)

    return parser


def get_workflow(logger):
    from nipype import logging as nlogging
    from multiprocessing import set_start_method, Process, Manager
    from ..utils.bids import validate_input_dir
    from .build_workflow import build_workflow
    set_start_method('forkserver')
    warnings.showwarning = _warn_redirect
    opts = get_parser().parse_args()

    exec_env = os.name

    # special variable set in the container
    # if os.getenv('IS_DOCKER_8395080871'):
    #     exec_env = 'singularity'
    #     cgroup = Path('/proc/1/cgroup')
    #     if cgroup.exists() and 'docker' in cgroup.read_text():
    #         exec_env = 'docker'
    #         if os.getenv('DOCKER_VERSION_8395080871'):
    #             exec_env = 'fmriprep-docker'

    sentry_sdk = None
    if not opts.notrack:
        import sentry_sdk
        from ..utils.sentry import sentry_setup
        sentry_setup(opts, exec_env)

    # Validate inputs
    if not opts.skip_bids_validation:
        print("Making sure the input data is BIDS compliant (warnings can be ignored in most "
              "cases).")
        validate_input_dir(exec_env, opts.bids_dir, opts.participant_label)

    # Retrieve logging level
    log_level = int(max(25 - 5 * opts.verbose_count, logging.DEBUG))
    # Set logging
    logger.setLevel(log_level)
    nlogging.getLogger('nipype.workflow').setLevel(log_level)
    nlogging.getLogger('nipype.interface').setLevel(log_level)
    nlogging.getLogger('nipype.utils').setLevel(log_level)

    # Call build_workflow(opts, retval)
    with Manager() as mgr:
        retval = mgr.dict()
        p = Process(target=build_workflow, args=(opts, retval))
        p.start()
        p.join()

        retcode = p.exitcode or retval.get('return_code', 0)

        bids_dir = Path(retval.get('bids_dir'))
        output_dir = Path(retval.get('output_dir'))
        work_dir = Path(retval.get('work_dir'))
        plugin_settings = retval.get('plugin_settings', None)
        subject_list = retval.get('subject_list', None)
        neuroHurst_wf = retval.get('workflow', None)
        run_uuid = retval.get('run_uuid', None)

    if opts.reports_only:
        sys.exit(int(retcode > 0))

    if opts.boilerplate:
        sys.exit(int(retcode > 0))

    if neuroHurst_wf and opts.write_graph:
        neuroHurst_wf.write_graph(graph2use="colored", format='svg', simple_form=True)

    retcode = retcode or int(neuroHurst_wf is None)
    if retcode != 0:
        sys.exit(retcode)

    # Check workflow for missing commands
    missing = check_deps(neuroHurst_wf)
    if missing:
        print("Cannot run neuroHurst. Missing dependencies:", file=sys.stderr)
        for iface, cmd in missing:
            print("\t{} (Interface: {})".format(cmd, iface))
        sys.exit(2)
    # Clean up master process before running workflow, which may create forks
    gc.collect()

    # Sentry tracking
    if not opts.notrack:
        from ..utils.sentry import start_ping
        start_ping(run_uuid, len(subject_list))

    return neuroHurst_wf, plugin_settings, opts, output_dir, work_dir, bids_dir, subject_list, run_uuid


from argparse import ArgumentParser
from bids import BIDSLayout
from nipype import Workflow
import sys
import os
from copy import deepcopy
from niworkflows.utils.misc import fix_multi_T1w_source_name

from nipype import __version__ as nipype_ver
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu

from niworkflows.engine.workflows import LiterateWorkflow as Workflow
from niworkflows.interfaces.bids import (
    BIDSInfo
)

from neuroHurst.workflows.datasink import init_datasink_wf
from ..utils.bids import collect_data, BIDSPlusDataGrabber

from ..workflows.dfa import init_dfa_workflow

from ..interfaces import SubjectSummary, AboutSummary, DerivativesDataSink
from ..__about__ import __version__


def init_base_wf(opts: ArgumentParser,
                 layout: BIDSLayout,
                 run_uuid: str,
                 subject_list: list,
                 work_dir: str,
                 output_dir: str):
    workflow = Workflow(name='neuroHurst_wf')
    workflow.base_dir = opts.work_dir

    reportlets_dir = os.path.join(opts.work_dir, 'repportlets')
    for subject_id in subject_list:
        single_subject_wf = init_single_subject_wf(
            opts=opts,
            layout=layout,
            run_uuid=run_uuid,
            work_dir=str(work_dir),
            output_dir=str(output_dir),
            name="single_subject_" + subject_id +"_wf",
            subject_id=subject_id,
            reportlets_dir=reportlets_dir,
        )

        single_subject_wf.config['execution']['crashdump_dir'] = (
            os.path.join(output_dir, "neuroHurst", "sub-" + subject_id, 'log', run_uuid)
        )
        for node in single_subject_wf._get_all_nodes():
            node.config = deepcopy(single_subject_wf.config)

        workflow.add_nodes([single_subject_wf])

    return workflow


def init_single_subject_wf(
            opts: ArgumentParser,
            layout: BIDSLayout,
            run_uuid: str,
            work_dir:str,
            output_dir:str,
            name:str,
            subject_id:str,
            reportlets_dir:str,
):
    import nilearn
    if name in ('single_subject_wf', 'single_subject_test_wf'):
        # for documentation purposes
        subject_data = {
            'bold': ['/completely/made/up/path/sub-01_task-nback_bold.nii.gz']
        }
    else:
        subject_data = collect_data(layout, subject_id, opts.source_format)[0]

    if not subject_data[opts.source_format]:
        raise Exception("No {data_format}. data found for participant {participant}. "
                        "All workflows require time series data.".format(
            data_format=opts.source_format,
            participant=subject_id)
        )

    workflow = Workflow(name=name)
    workflow.__desc__ = """
    Results included in this manuscript come from processing
    performed using *neuroHurst* {neuroHurst_ver}
    (@neurohurst1; @neurohurst; RRID:SCR_016216),
    which is based on *Nipype* {nipype_ver}
    (@nipype1; @nipype2; RRID:SCR_002502).
    """.format(neuroHurst_ver=__version__, nipype_ver=nipype_ver)
    workflow.__postdesc__ = """
    Many internal operations of *neuroHurst* use
    *Nilearn* {nilearn_ver} [@nilearn, RRID:SCR_001362].
    ### Copyright Waiver
    The above boilerplate text was automatically generated by fMRIPrep
    with the express intention that users should copy and paste this
    text into their manuscripts *unchanged*.
    It is released under the [CC0]\
    (https://creativecommons.org/publicdomain/zero/1.0/) license.
    ### References
    """.format(nilearn_ver=nilearn.version.__version__)

    inputnode = pe.Node(niu.IdentityInterface(fields=['subjects_dir']),
                        name='inputnode')

    require_masks = opts.source_format == 'bold'
    bidssrc = pe.Node(BIDSPlusDataGrabber(subject_data=subject_data, require_masks=require_masks),
                      name='bidssrc')

    bids_info = pe.Node(BIDSInfo(
        bids_dir=layout.root, bids_validate=False), name='bids_info')

    summary = pe.Node(SubjectSummary(),
        name='summary', run_without_submitting=True)

    about = pe.Node(AboutSummary(version=__version__,
                                 command=' '.join(sys.argv)),
                    name='about', run_without_submitting=True)

    ds_report_summary = pe.Node(
        DerivativesDataSink(base_directory=reportlets_dir,
                            desc='summary', keep_dtype=True),
        name='ds_report_summary', run_without_submitting=True)

    ds_report_about = pe.Node(
        DerivativesDataSink(base_directory=reportlets_dir,
                            desc='about', keep_dtype=True),
        name='ds_report_about', run_without_submitting=True)

    # Preprocessing of T1w (includes registration to MNI)

    workflow.connect([
        (bidssrc, bids_info, [('bold', 'in_file')]),
        (inputnode, summary, [('subjects_dir', 'subjects_dir')]),
        (bidssrc, summary, [('t1w', 't1w'),
                            ('t2w', 't2w'),
                            ('bold', 'bold')]),
        (bids_info, summary, [('subject', 'subject_id')]),
        (bidssrc, ds_report_summary, [('bold', 'source_file')]),
        (summary, ds_report_summary, [('out_report', 'in_file')]),
        (bidssrc, ds_report_about, [('bold', 'source_file')]),
        (about, ds_report_about, [('out_report', 'in_file')]),
    ])

    # Overwrite ``out_path_base`` of smriprep's DataSinks
    for node in workflow.list_node_names():
        if node.split('.')[-1].startswith('ds_'):
            workflow.get_node(node).interface.out_path_base = 'neuroHurst'

    for i in range(len(subject_data['bold'])):
        dfa_wf = init_dfa_workflow(
            bold=subject_data['bold'][i],
            brainmask=subject_data['mask'][i] if subject_data.keys().__contains__('mask')else None,
            csv=subject_data['csv'][i] if subject_data.keys().__contains__('csv') else None,
            min_freq=opts.minimum_frequency,
            max_freq=opts.maximum_frequency,
            drop_vols=opts.skip_vols,
            name='dfa_%d_wf' % i
        )
        workflow.connect([
            (inputnode, dfa_wf, [('subjects_dir', 'inputnode.subjects_dir')]),
        ])

        outputs_wf = init_datasink_wf(bids_root=str(layout.root), output_dir=str(opts.output_dir), name='ds_%d_wf' % i)

        workflow.connect([(dfa_wf,outputs_wf,[('inputnode.bold','inputnode.source_file')])])
        # outputs_wf.inputs.source_file = subject_data['csv'][i] if subject_data.__contains__('mask') else subject_data['bold'][i]

        workflow.connect([(dfa_wf, outputs_wf, [
            ('outputnode.hurst', 'inputnode.dfa_h'),
            ('outputnode.rsquared', 'inputnode.dfa_ci'),
            ('outputnode.confidence_intervals', 'inputnode.dfa_rsquared')
        ])])

    return workflow


def _prefix(subid):
    if subid.startswith('sub-'):
        return subid
    return '-'.join(('sub', subid))


def _pop(inlist):
    if isinstance(inlist, (list, tuple)):
        return inlist[0]
    return inlist


from nipype.pipeline import engine as pe
from nipype.interfaces import (
    utility as niu,

)
from niworkflows.engine.workflows import LiterateWorkflow as Workflow
from ..interfaces import DFA, DerivativesDataSink


def init_dfa_workflow(bold, brainmask, csv, min_freq=0.01, max_freq=0.1, drop_vols=5, name='dfa_wf'):#bids_root, omp_nthreads, output_dir, output_format, reportlets_dir, debug=False, name='dfa_wf'):

    workflow = Workflow(name=name)
    desc = """Fractal Scaling via Detrended Fluctuation Analysis
    : """
    # desc += """\
    # A total of {num_t1w} T1-weighted (T1w) images were found within the input
    # BIDS dataset.
    # All of them were corrected for intensity non-uniformity (INU)
    # """ if num_t1w > 1 else """\
    # The T1-weighted (T1w) image was corrected for intensity non-uniformity (INU)
    # """
    # desc += """\
    # with `N4BiasFieldCorrection` [@n4], distributed with ANTs {ants_ver} \
    # [@ants, RRID:SCR_004757]"""
    # desc += '.\n' if num_t1w > 1 else ", and used as T1w-reference throughout the workflow.\n"
    #
    # desc += """\
    # The T1w-reference was then skull-stripped with a *Nipype* implementation of
    # the `antsBrainExtraction.sh` workflow (from ANTs), using {skullstrip_tpl}
    # as target template.
    # Brain tissue segmentation of cerebrospinal fluid (CSF),
    # white-matter (WM) and gray-matter (GM) was performed on
    # the brain-extracted T1w using `fast` [FSL {fsl_ver}, RRID:SCR_002823,
    # @fsl_fast].
    # """
    #
    # workflow.__desc__ = desc.format(
    #     ants_ver=ANTsInfo.version() or '(version unknown)',
    #     fsl_ver=fsl.FAST().version or '(version unknown)',
    #     num_t1w=num_t1w,
    #     skullstrip_tpl=skull_strip_template[0],
    # )

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['bold','csv', 'max_frequency', 'min_frequency', 'drop_vols', 'brainmask', 'subjects_dir']),
        name='inputnode')

    inputnode.inputs.bold = bold
    inputnode.inputs.csv = csv
    inputnode.inputs.max_frequency = max_freq
    inputnode.inputs.min_frequency = min_freq
    inputnode.inputs.brainmask = brainmask
    inputnode.inputs.drop_vols = drop_vols

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['rsquared', 'confidence_intervals', 'hurst']),
        name='outputnode')

    dfanode = pe.Node(DFA(), name='dfa')

    workflow.connect([
        (inputnode, dfanode, [('bold', 'bold')]),
        (inputnode, dfanode, [('max_frequency', 'max_frequency')]),
        (inputnode, dfanode, [('min_frequency', 'min_frequency')]),
        (inputnode, dfanode, [('drop_vols', 'drop_vols')]),
        # (dfanode, outputnode, [('out_report', 'out_report')]),
        (dfanode, outputnode, [('hurst', 'hurst')]),
        (dfanode, outputnode, [('confidence_intervals', 'confidence_intervals')]),
        (dfanode, outputnode, [('rsquared', 'rsquared')]),
    ])

    if csv is not None:
        workflow.connect([
            (inputnode, dfanode, [('csv', 'csv')]),
        ])
    if brainmask is not None:
        workflow.connect([
            (inputnode, dfanode, [('brainmask', 'brainmask')]),
        ])








    # ds_report_summary = pe.Node(
    #     DerivativesDataSink(desc='summary', keep_dtype=True),
    #     name='ds_report_summary', run_without_submitting=True,
    #     mem_gb=1)

    # summary = pe.Node(
    #     FunctionalSummary(
    #         slice_timing=run_stc,
    #         registration=('FSL', 'FreeSurfer')[freesurfer],
    #         registration_dof=bold2t1w_dof,
    #         pe_direction=metadata.get("PhaseEncodingDirection"),
    #         tr=metadata.get("RepetitionTime")),
    #     name='summary', mem_gb=DEFAULT_MEMORY_MIN_GB, run_without_submitting=True)
    # summary.inputs.dummy_scans = dummy_scans

    # workflow.connect([
    #     (summary, ds_report_summary, [('out_report', 'in_file')]),
    #     (bold_reference_wf, ds_report_validation, [
    #         ('outputnode.validation_report', 'in_file')]),
    # ])
    return workflow


import re
from multiprocessing import cpu_count
from pathlib import Path

from bids import BIDSLayout


def build_workflow(opts, retval):
    from nipype import logging as nlogging, config as ncfg
    from niworkflows.utils.bids import collect_participants
    from niworkflows.reports import generate_reports
    from ..__about__ import __version__
    from time import strftime
    import uuid
    from ..workflows.base import init_base_wf

    build_log = nlogging.getLogger('nipype.workflow')

    INIT_MSG = """
    #     Running neuroHurst version {version}:
    #       * BIDS dataset path: {bids_dir}.
    #       * Participant list: {subject_list}.
    #       * Run identifier: {uuid}.
    #     """.format

    pass

    bids_dir = opts.bids_dir.resolve()
    output_dir = opts.output_dir.resolve()
    work_dir = opts.work_dir.resolve()

    retval['return_code'] = 1
    retval['workflow'] = None
    retval['bids_dir'] = str(bids_dir)
    retval['output_dir'] = str(output_dir)
    retval['work_dir'] = str(work_dir)

    if output_dir == bids_dir:
        build_log.error(
            'The selected output folder is the same as the input BIDS folder. '
            'Please modify the output path (suggestion: %s).',
            bids_dir / 'derivatives' / ('neuroHurst-%s' % __version__.split('+')[0]))
        retval['return_code'] = 1
        return retval

    # Set up some instrumental utilities
    run_uuid = '%s_%s' % (strftime('%Y%m%d-%H%M%S'), uuid.uuid4())
    retval['run_uuid'] = run_uuid

    # First check that bids_dir looks like a BIDS folder
    layout = BIDSLayout(str(bids_dir), validate=False,
                        ignore=("code", "stimuli", "sourcedata", "models",
                                "derivatives", re.compile(r'^\.')))
    subject_list = collect_participants(
        layout, participant_label=opts.participant_label)
    retval['subject_list'] = subject_list

    # Load base plugin_settings from file if --use-plugin
    if opts.use_plugin is not None:
        from yaml import load as loadyml
        with open(opts.use_plugin) as f:
            plugin_settings = loadyml(f)
        plugin_settings.setdefault('plugin_args', {})
    else:
        # Defaults
        plugin_settings = {
            'plugin': 'MultiProc',
            'plugin_args': {
                'raise_insufficient': False,
                'maxtasksperchild': 1,
            }
        }

    # Resource management options
    # Note that we're making strong assumptions about valid plugin args
    # This may need to be revisited if people try to use batch plugins
    nthreads = plugin_settings['plugin_args'].get('n_procs')
    # Permit overriding plugin config with specific CLI options
    if nthreads is None or opts.nthreads is not None:
        nthreads = opts.nthreads
        if nthreads is None or nthreads < 1:
            nthreads = cpu_count()
        plugin_settings['plugin_args']['n_procs'] = nthreads

    if opts.mem_mb:
        plugin_settings['plugin_args']['memory_gb'] = opts.mem_mb / 1024

    omp_nthreads = opts.omp_nthreads
    if omp_nthreads == 0:
        omp_nthreads = min(nthreads - 1 if nthreads > 1 else cpu_count(), 8)

    if 1 < nthreads < omp_nthreads:
        build_log.warning(
            'Per-process threads (--omp-nthreads=%d) exceed total '
            'threads (--nthreads/--n_cpus=%d)', omp_nthreads, nthreads)
    retval['plugin_settings'] = plugin_settings

    # Set up directories
    log_dir = output_dir / 'neuroHurst' / 'logs'
    # Check and create output and working directories
    output_dir.mkdir(exist_ok=True, parents=True)
    log_dir.mkdir(exist_ok=True, parents=True)
    work_dir.mkdir(exist_ok=True, parents=True)

    # Nipype config (logs and execution)
    ncfg.update_config({
        'logging': {
            'log_directory': str(log_dir),
            'log_to_file': True
        },
        'execution': {
            'crashdump_dir': str(log_dir),
            'crashfile_format': 'txt',
            'get_linked_libs': False,
            'stop_on_first_crash': opts.stop_on_first_crash,
        },
        'monitoring': {
            'enabled': opts.resource_monitor,
            'sample_frequency': '0.5',
            'summary_append': True,
        }
    })

    if opts.resource_monitor:
        ncfg.enable_resource_monitor()

    # Called with reports only
    if opts.reports_only:
        build_log.log(25, 'Running --reports-only on participants %s', ', '.join(subject_list))
        if opts.run_uuid is not None:
            run_uuid = opts.run_uuid
            retval['run_uuid'] = run_uuid
        retval['return_code'] = generate_reports(
            subject_list, output_dir, work_dir, run_uuid,
            packagename='neuroHurst')
        return retval

    # Build main workflow
    build_log.log(25, INIT_MSG(
        version=__version__,
        bids_dir=bids_dir,
        subject_list=subject_list,
        uuid=run_uuid)
                  )

    retval['workflow'] = init_base_wf(
        opts=opts,
        layout=layout,
        run_uuid=run_uuid,
        subject_list=subject_list,
        work_dir=str(work_dir),
        output_dir=str(output_dir),
    )
    retval['return_code'] = 0

    logs_path = Path(output_dir) / 'neuroHurst' / 'logs'
    boilerplate = retval['workflow'].visit_desc()

    if boilerplate:
        citation_files = {
            ext: logs_path / ('CITATION.%s' % ext)
            for ext in ('bib', 'tex', 'md', 'html')
        }
        # To please git-annex users and also to guarantee consistency
        # among different renderings of the same file, first remove any
        # existing one
        for citation_file in citation_files.values():
            try:
                citation_file.unlink()
            except FileNotFoundError:
                pass

        citation_files['md'].write_text(boilerplate)
        build_log.log(25, 'Works derived from this neuroHurst execution should '
                          'include the following boilerplate:\n\n%s', boilerplate)
    return retval



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Version CLI helpers."""

from pathlib import Path
from datetime import datetime
import requests
from .. import __version__

RELEASE_EXPIRY_DAYS = 14
DATE_FMT = '%Y%m%d'


def check_latest():
    """Determine whether this is the latest version."""
    from packaging.version import Version, InvalidVersion

    latest = None
    date = None
    outdated = None
    cachefile = Path.home() / '.cache' / 'neuroHurst' / 'latest'
    try:
        cachefile.parent.mkdir(parents=True, exist_ok=True)
    except OSError:
        cachefile = None

    if cachefile and cachefile.exists():
        try:
            latest, date = cachefile.read_text().split('|')
        except Exception:
            pass
        else:
            try:
                latest = Version(latest)
                date = datetime.strptime(date, DATE_FMT)
            except (InvalidVersion, ValueError):
                latest = None
            else:
                if abs((datetime.now() - date).days) > RELEASE_EXPIRY_DAYS:
                    outdated = True

    if latest is None or outdated is True:
        try:
            response = requests.get(url='https://pypi.org/pypi/neuroHurst/json', timeout=1.0)
        except Exception:
            response = None

        if response and response.status_code == 200:
            versions = [Version(rel) for rel in response.json()['releases'].keys()]
            versions = [rel for rel in versions if not rel.is_prerelease]
            if versions:
                latest = sorted(versions)[-1]
        else:
            latest = None

    if cachefile is not None and latest is not None:
        try:
            cachefile.write_text('|'.join(('%s' % latest, datetime.now().strftime(DATE_FMT))))
        except Exception:
            pass

    return latest


def is_flagged():
    """Check whether current version is flagged."""
    # https://raw.githubusercontent.com/enlberman/neuroHurst/master/.versions.json
    flagged = tuple()
    try:
        response = requests.get(url="""\
https://raw.githubusercontent.com/enlberman/neuroHurst/master/.versions.json""", timeout=1.0)
    except Exception:
        response = None

    if response and response.status_code == 200:
        flagged = response.json().get('flagged', {}) or {}

    if __version__ in flagged:
        return True, flagged[__version__]

    return False, None



#!/usr/bin/env python
import sys
from setuptools import setup
import versioneer

# Give setuptools a hint to complain if it's too old a version
# 40.8.0 allows us to put most metadata in setup.cfg
# Should match pyproject.toml
SETUP_REQUIRES = ['setuptools >= 40.8']
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []

if __name__ == '__main__':
    setup(name="neuroHurst",
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass(),
          setup_requires=SETUP_REQUIRES, install_requires=['smriprep', 'packaging', 'bids'])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Interfaces to generate reportlets
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import os
import time
import re

from collections import Counter
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, Directory, InputMultiObject, Str, isdefined,
    SimpleInterface)
from nipype.interfaces import freesurfer as fs
from niworkflows.utils.bids import BIDS_NAME


SUBJECT_TEMPLATE = """\
\t<ul class="elem-desc">
\t\t<li>Subject ID: {subject_id}</li>
\t\t<li>Structural images: {n_t1s:d} T1-weighted {t2w}</li>
\t\t<li>Functional series: {n_bold:d}</li>
{tasks}
\t\t<li>Other series: {n_other:d}</li>
{tasks}
\t</ul>
"""

ABOUT_TEMPLATE = """\t<ul>
\t\t<li>neuroHurst version: {version}</li>
\t\t<li>neuroHurrst command: <code>{command}</code></li>
\t\t<li>Date preprocessed: {date}</li>
\t</ul>
</div>
"""


class SummaryOutputSpec(TraitedSpec):
    out_report = File(exists=True, desc='HTML segment containing summary')


class SummaryInterface(SimpleInterface):
    output_spec = SummaryOutputSpec

    def _run_interface(self, runtime):
        segment = self._generate_segment()
        fname = os.path.join(runtime.cwd, 'report.html')
        with open(fname, 'w') as fobj:
            fobj.write(segment)
        self._results['out_report'] = fname
        return runtime

    def _generate_segment(self):
        raise NotImplementedError


class SubjectSummaryInputSpec(BaseInterfaceInputSpec):
    t1w = InputMultiObject(File(exists=True), desc='T1w structural images')
    t2w = InputMultiObject(File(exists=True), desc='T2w structural images')
    subjects_dir = Directory(desc='FreeSurfer subjects directory')
    subject_id = Str(desc='Subject ID')
    bold = InputMultiObject(traits.Either(
        File(exists=True), traits.List(File(exists=True))),
        desc='BOLD functional series')
    other_ts = InputMultiObject(traits.Either(
        File(exists=True), traits.List(File(exists=True))),
        desc='Other series')


class SubjectSummaryOutputSpec(SummaryOutputSpec):
    # This exists to ensure that the summary is run prior to the first ReconAll
    # call, allowing a determination whether there is a pre-existing directory
    subject_id = Str(desc='FreeSurfer subject ID')


class SubjectSummary(SummaryInterface):
    input_spec = SubjectSummaryInputSpec
    output_spec = SubjectSummaryOutputSpec

    def _run_interface(self, runtime):
        if isdefined(self.inputs.subject_id):
            self._results['subject_id'] = self.inputs.subject_id
        return super(SubjectSummary, self)._run_interface(runtime)

    def _generate_segment(self):
        t2w_seg = ''
        if self.inputs.t2w:
            t2w_seg = '(+ {:d} T2-weighted)'.format(len(self.inputs.t2w))

        # Add list of tasks with number of runs
        bold_series = self.inputs.bold if isdefined(self.inputs.bold) else []
        bold_series = [s[0] if isinstance(s, list) else s for s in bold_series]

        other_series = self.inputs.other_ts if isdefined(self.inputs.other_ts) else []
        other_series = [s[0] if isinstance(s, list) else s for s in other_series]

        counts = Counter(BIDS_NAME.search(series).groupdict()['task_id'][5:]
                         for series in bold_series)

        if not bold_series:
            counts = Counter(BIDS_NAME.search(series).groupdict()['task_id'][5:]
                             for series in other_series)

        tasks = ''
        if counts:
            header = '\t\t<ul class="elem-desc">'
            footer = '\t\t</ul>'
            lines = ['\t\t\t<li>Task: {task_id} ({n_runs:d} run{s})</li>'.format(
                     task_id=task_id, n_runs=n_runs, s='' if n_runs == 1 else 's')
                     for task_id, n_runs in sorted(counts.items())]
            tasks = '\n'.join([header] + lines + [footer])

        return SUBJECT_TEMPLATE.format(
            subject_id=self.inputs.subject_id,
            n_t1s=len(self.inputs.t1w),
            t2w=t2w_seg,
            n_bold=len(bold_series),
            n_other=len(other_series),
            tasks=tasks)


class FunctionalSummaryInputSpec(BaseInterfaceInputSpec):
    slice_timing = traits.Enum(False, True, 'TooShort', usedefault=True,
                               desc='Slice timing correction used')
    distortion_correction = traits.Str(desc='Susceptibility distortion correction method',
                                       mandatory=True)
    pe_direction = traits.Enum(None, 'i', 'i-', 'j', 'j-', mandatory=True,
                               desc='Phase-encoding direction detected')
    registration = traits.Enum('FSL', 'FreeSurfer', mandatory=True,
                               desc='Functional/anatomical registration method')
    fallback = traits.Bool(desc='Boundary-based registration rejected')
    registration_dof = traits.Enum(6, 9, 12, desc='Registration degrees of freedom',
                                   mandatory=True)
    confounds_file = File(exists=True, desc='Confounds file')
    tr = traits.Float(desc='Repetition time', mandatory=True)
    dummy_scans = traits.Either(traits.Int(), None, desc='number of dummy scans specified by user')
    algo_dummy_scans = traits.Int(desc='number of dummy scans determined by algorithm')


class AboutSummaryInputSpec(BaseInterfaceInputSpec):
    version = Str(desc='neuroHurst version')
    command = Str(desc='neuroHurst command')
    # Date not included - update timestamp only if version or command changes


class AboutSummary(SummaryInterface):
    input_spec = AboutSummaryInputSpec

    def _generate_segment(self):
        return ABOUT_TEMPLATE.format(version=self.inputs.version,
                                     command=self.inputs.command,
                                     date=time.strftime("%Y-%m-%d %H:%M:%S %z"))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Utilities to handle BIDS inputs
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""
import os
import sys
import json
from pathlib import Path

from bids import BIDSLayout
from nipype import logging
from nipype.interfaces.base import TraitedSpec, OutputMultiObject, SimpleInterface
from niworkflows.interfaces.bids import BIDSDataGrabberOutputSpec, BIDSDataGrabber, BIDSDataGrabberInputSpec, \
    BIDSInfoInputSpec

LOGGER = logging.getLogger('nipype.interface')


def write_derivative_description(bids_dir, deriv_dir):
    from ..__about__ import __version__, __url__, DOWNLOAD_URL

    bids_dir = Path(bids_dir)
    deriv_dir = Path(deriv_dir)
    desc = {
        'Name': 'neuroHurst - fractal scaling for neuroscience',
        'BIDSVersion': '1.1.1',
        'PipelineDescription': {
            'Name': 'neuroHurst',
            'Version': __version__,
            'CodeURL': DOWNLOAD_URL,
        },
        'CodeURL': __url__,
        # 'HowToAcknowledge':
        #     'Please cite our paper (https://doi.org/10.1038/s41592-018-0235-4), '
        #     'and include the generated citation boilerplate within the Methods '
        #     'section of the text.',
    }

    # Keys that can only be set by environment
    if 'NEUROHURST_DOCKER_TAG' in os.environ:
        desc['DockerHubContainerTag'] = os.environ['NEUROHURST_DOCKER_TAG']
    if 'NEUROHURST_SINGULARITY_URL' in os.environ:
        singularity_url = os.environ['NEUROHURST_SINGULARITY_URL']
        desc['SingularityContainerURL'] = singularity_url

        singularity_md5 = _get_shub_version(singularity_url)
        if singularity_md5 and singularity_md5 is not NotImplemented:
            desc['SingularityContainerMD5'] = _get_shub_version(singularity_url)

    # Keys deriving from source dataset
    orig_desc = {}
    fname = bids_dir / 'dataset_description.json'
    if fname.exists():
        with fname.open() as fobj:
            orig_desc = json.load(fobj)

    if 'DatasetDOI' in orig_desc:
        desc['SourceDatasetsURLs'] = ['https://doi.org/{}'.format(
            orig_desc['DatasetDOI'])]
    if 'License' in orig_desc:
        desc['License'] = orig_desc['License']

    with (deriv_dir / 'dataset_description.json').open('w') as fobj:
        json.dump(desc, fobj, indent=4)


def validate_input_dir(exec_env, bids_dir, participant_label):
    # Ignore issues and warnings that should not influence FMRIPREP
    import tempfile
    import subprocess
    validator_config_dict = {
        "ignore": [
            "EVENTS_COLUMN_ONSET",
            "EVENTS_COLUMN_DURATION",
            "TSV_EQUAL_ROWS",
            "TSV_EMPTY_CELL",
            "TSV_IMPROPER_NA",
            "VOLUME_COUNT_MISMATCH",
            "BVAL_MULTIPLE_ROWS",
            "BVEC_NUMBER_ROWS",
            "DWI_MISSING_BVAL",
            "INCONSISTENT_SUBJECTS",
            "INCONSISTENT_PARAMETERS",
            "BVEC_ROW_LENGTH",
            "B_FILE",
            "PARTICIPANT_ID_COLUMN",
            "PARTICIPANT_ID_MISMATCH",
            "TASK_NAME_MUST_DEFINE",
            "PHENOTYPE_SUBJECTS_MISSING",
            "STIMULUS_FILE_MISSING",
            "DWI_MISSING_BVEC",
            "EVENTS_TSV_MISSING",
            "TSV_IMPROPER_NA",
            "ACQTIME_FMT",
            "Participants age 89 or higher",
            "DATASET_DESCRIPTION_JSON_MISSING",
            "FILENAME_COLUMN",
            "WRONG_NEW_LINE",
            "MISSING_TSV_COLUMN_CHANNELS",
            "MISSING_TSV_COLUMN_IEEG_CHANNELS",
            "MISSING_TSV_COLUMN_IEEG_ELECTRODES",
            "UNUSED_STIMULUS",
            "CHANNELS_COLUMN_SFREQ",
            "CHANNELS_COLUMN_LOWCUT",
            "CHANNELS_COLUMN_HIGHCUT",
            "CHANNELS_COLUMN_NOTCH",
            "CUSTOM_COLUMN_WITHOUT_DESCRIPTION",
            "ACQTIME_FMT",
            "SUSPICIOUSLY_LONG_EVENT_DESIGN",
            "SUSPICIOUSLY_SHORT_EVENT_DESIGN",
            "MALFORMED_BVEC",
            "MALFORMED_BVAL",
            "MISSING_TSV_COLUMN_EEG_ELECTRODES",
            "MISSING_SESSION"
        ],
        "error": ["NO_T1W"],
        "ignoredFiles": ['/dataset_description.json', '/participants.tsv']
    }
    # Limit validation only to data from requested participants
    if participant_label:
        all_subs = set([s.name[4:] for s in bids_dir.glob('sub-*')])
        selected_subs = set([s[4:] if s.startswith('sub-') else s
                             for s in participant_label])
        bad_labels = selected_subs.difference(all_subs)
        if bad_labels:
            error_msg = 'Data for requested participant(s) label(s) not found. Could ' \
                        'not find data for participant(s): %s. Please verify the requested ' \
                        'participant labels.'
            if exec_env == 'docker':
                error_msg += ' This error can be caused by the input data not being ' \
                             'accessible inside the docker container. Please make sure all ' \
                             'volumes are mounted properly (see https://docs.docker.com/' \
                             'engine/reference/commandline/run/#mount-volume--v---read-only)'
            if exec_env == 'singularity':
                error_msg += ' This error can be caused by the input data not being ' \
                             'accessible inside the singularity container. Please make sure ' \
                             'all paths are mapped properly (see https://www.sylabs.io/' \
                             'guides/3.0/user-guide/bind_paths_and_mounts.html)'
            raise RuntimeError(error_msg % ','.join(bad_labels))

        ignored_subs = all_subs.difference(selected_subs)
        if ignored_subs:
            for sub in ignored_subs:
                validator_config_dict["ignoredFiles"].append("/sub-%s/**" % sub)
    with tempfile.NamedTemporaryFile('w+') as temp:
        temp.write(json.dumps(validator_config_dict))
        temp.flush()
        try:
            subprocess.check_call(['bids-validator', bids_dir, '-c', temp.name])
        except FileNotFoundError:
            print("bids-validator does not appear to be installed", file=sys.stderr)


def _get_shub_version(singularity_url):
    return NotImplemented


def collect_data(layout: BIDSLayout, subject_id, other_format: str = None):
    from niworkflows.utils.bids import collect_data
    standard_search, layout = collect_data(layout, subject_id)
    masks = [file.replace('preproc_bold','brain_mask') for file in standard_search['bold']]
    standard_search['mask'] = masks
    return standard_search, layout


class BIDSPlusDataGrabberOutputSpec(BIDSDataGrabberOutputSpec):
    csv = OutputMultiObject(desc='output csv')
    mask = OutputMultiObject(desc='output mask')


class BIDSPlusDataGrabber(BIDSDataGrabber):
    """
    Collect files from a BIDS directory structure plus support for other derivatives
    """

    input_spec = BIDSDataGrabberInputSpec
    output_spec = BIDSPlusDataGrabberOutputSpec
    _require_funcs = True
    _require_masks = False
    _require_anat = False

    def __init__(self, *args, **kwargs):
        # anat_only = kwargs.pop('anat_only')
        # require_t1w = kwargs.pop('require_t1w')
        require_masks = kwargs.pop('require_masks')
        super(BIDSDataGrabber, self).__init__(*args, **kwargs)
        # if anat_only is not None:
        #     self._require_funcs = not anat_only
        # if require_t1w is not None:
        #     self._require_anat = require_t1w
        if require_masks is not None:
            self._require_funcs = require_masks

    def _run_interface(self, runtime):
        bids_dict = self.inputs.subject_data

        self._results['out_dict'] = bids_dict
        self._results.update(bids_dict)

        if not bids_dict['t1w'] and self._require_anat:
            raise FileNotFoundError('No T1w images found for subject sub-{}'.format(
                self.inputs.subject_id))

        if self._require_funcs and not bids_dict['bold']:
            raise FileNotFoundError('No functional images found for subject sub-{}'.format(
                self.inputs.subject_id))

        if self._require_masks and not bids_dict['mask']:
            raise FileNotFoundError('No functional mask images found for subject sub-{}'.format(
                self.inputs.subject_id))

        for imtype in ['bold', 't2w', 'flair', 'fmap', 'sbref', 'roi', 'csv', 'mask']:
            if not bids_dict.keys().__contains__(imtype):
                LOGGER.warning('No "%s" images found for sub-%s',
                               imtype, self.inputs.subject_id)

        return runtime





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Stripped out routines for Sentry"""
import os
from pathlib import Path
import re
from niworkflows.utils.misc import read_crashfile
import sentry_sdk

CHUNK_SIZE = 16384
# Group common events with pre specified fingerprints
KNOWN_ERRORS = {
    'permission-denied': [
        "PermissionError: [Errno 13] Permission denied"
    ],
    'memory-error': [
        "MemoryError",
        "Cannot allocate memory",
        "Return code: 134",
    ],

    'no-disk-space': [
        "[Errno 28] No space left on device",
        "[Errno 122] Disk quota exceeded"
    ],
    'segfault': [
        "Segmentation Fault",
        "Segfault",
        "Return code: 139",
    ],
    'potential-race-condition': [
        "[Errno 39] Directory not empty",
        "_unfinished.json",
    ],
    'keyboard-interrupt': [
        "KeyboardInterrupt",
    ],
}


def start_ping(run_uuid, npart):
    with sentry_sdk.configure_scope() as scope:
        if run_uuid:
            scope.set_tag('run_uuid', run_uuid)
        scope.set_tag('npart', npart)
    sentry_sdk.add_breadcrumb(message='neuroHurst started', level='info')
    sentry_sdk.capture_message('neuroHurst started', level='info')


def sentry_setup(opts, exec_env):
    from os import cpu_count
    import psutil
    import hashlib
    from ..__about__ import __version__

    environment = "prod"
    release = __version__
    if not __version__:
        environment = "dev"
        release = "dev"
    elif int(os.getenv('neuroHurst_DEV', '0')) or ('+' in __version__):
        environment = "dev"

    sentry_sdk.init("https://d5a16b0c38d84d1584dfc93b9fb1ade6@sentry.io/1137693",
                    release=release,
                    environment=environment,
                    before_send=before_send)
    with sentry_sdk.configure_scope() as scope:
        scope.set_tag('exec_env', exec_env)

        # if exec_env == 'fmriprep-docker':
        #     scope.set_tag('docker_version', os.getenv('DOCKER_VERSION_8395080871'))

        dset_desc_path = opts.bids_dir / 'dataset_description.json'
        if dset_desc_path.exists():
            desc_content = dset_desc_path.read_bytes()
            scope.set_tag('dset_desc_sha256', hashlib.sha256(desc_content).hexdigest())

        free_mem_at_start = round(psutil.virtual_memory().free / 1024**3, 1)
        scope.set_tag('free_mem_at_start', free_mem_at_start)
        scope.set_tag('cpu_count', cpu_count())

        # Memory policy may have a large effect on types of errors experienced
        overcommit_memory = Path('/proc/sys/vm/overcommit_memory')
        if overcommit_memory.exists():
            policy = {'0': 'heuristic',
                      '1': 'always',
                      '2': 'never'}.get(overcommit_memory.read_text().strip(), 'unknown')
            scope.set_tag('overcommit_memory', policy)
            if policy == 'never':
                overcommit_kbytes = Path('/proc/sys/vm/overcommit_memory')
                kb = overcommit_kbytes.read_text().strip()
                if kb != '0':
                    limit = '{}kB'.format(kb)
                else:
                    overcommit_ratio = Path('/proc/sys/vm/overcommit_ratio')
                    limit = '{}%'.format(overcommit_ratio.read_text().strip())
                scope.set_tag('overcommit_limit', limit)
            else:
                scope.set_tag('overcommit_limit', 'n/a')
        else:
            scope.set_tag('overcommit_memory', 'n/a')
            scope.set_tag('overcommit_limit', 'n/a')

        for k, v in vars(opts).items():
            scope.set_tag(k, v)


def process_crashfile(crashfile):
    """Parse the contents of a crashfile and submit sentry messages"""
    crash_info = read_crashfile(str(crashfile))
    with sentry_sdk.push_scope() as scope:
        scope.level = 'fatal'

        # Extract node name
        node_name = crash_info.pop('node').split('.')[-1]
        scope.set_tag("node_name", node_name)

        # Massage the traceback, extract the gist
        traceback = crash_info.pop('traceback')
        # last line is probably most informative summary
        gist = traceback.splitlines()[-1]
        exception_text_start = 1
        for line in traceback.splitlines()[1:]:
            if not line[0].isspace():
                break
            exception_text_start += 1

        exception_text = '\n'.join(
            traceback.splitlines()[exception_text_start:])

        # Extract inputs, if present
        inputs = crash_info.pop('inputs', None)
        if inputs:
            scope.set_extra('inputs', dict(inputs))

        # Extract any other possible metadata in the crash file
        for k, v in crash_info.items():
            strv = list(_chunks(str(v)))
            if len(strv) == 1:
                scope.set_extra(k, strv[0])
            else:
                for i, chunk in enumerate(strv):
                    scope.set_extra('%s_%02d' % (k, i), chunk)

        fingerprint = ''
        issue_title = '{}: {}'.format(node_name, gist)
        for new_fingerprint, error_snippets in KNOWN_ERRORS.items():
            for error_snippet in error_snippets:
                if error_snippet in traceback:
                    fingerprint = new_fingerprint
                    issue_title = new_fingerprint
                    break
            if fingerprint:
                break

        message = issue_title + '\n\n'
        message += exception_text[-(8192 - len(message)):]
        if fingerprint:
            sentry_sdk.add_breadcrumb(message=fingerprint, level='fatal')
        else:
            # remove file paths
            fingerprint = re.sub(r"(/[^/ ]*)+/?", '', message)
            # remove words containing numbers
            fingerprint = re.sub(r"([a-zA-Z]*[0-9]+[a-zA-Z]*)+", '', fingerprint)
            # adding the return code if it exists
            for line in message.splitlines():
                if line.startswith("Return code"):
                    fingerprint += line
                    break

        scope.fingerprint = [fingerprint]
        sentry_sdk.capture_message(message, 'fatal')


def before_send(event, hints):
    # Filtering log messages about crashed nodes
    if 'logentry' in event and 'message' in event['logentry']:
        msg = event['logentry']['message']
        if msg.startswith("could not run node:"):
            return None
        if msg.startswith("Saving crash info to "):
            return None
        if re.match("Node .+ failed to run on host .+", msg):
            return None

    if 'breadcrumbs' in event and isinstance(event['breadcrumbs'], list):
        fingerprints_to_propagate = ['no-disk-space', 'memory-error', 'permission-denied',
                                     'keyboard-interrupt']
        for bc in event['breadcrumbs']:
            msg = bc.get('message', 'empty-msg')
            if msg in fingerprints_to_propagate:
                event['fingerprint'] = [msg]
                break

    return event


def _chunks(string, length=CHUNK_SIZE):
    """
    Splits a string into smaller chunks
    >>> list(_chunks('some longer string.', length=3))
    ['som', 'e l', 'ong', 'er ', 'str', 'ing', '.']
    """
    return (string[i:i + length]
            for i in range(0, len(string), length))